#
# timeInStatus.R, 12 Jun 22
#
# Data from:
# https://github.com/robert-krasinski/JiraAndPythonForManagers
#

library("bizdays")
# library("msm")
library("plyr")
library("survival")

# Build observation history for a one status
mk_status_hist=function(df_S)
{
# Build observation history for one row, i.e., unique value of X
   mk_X_times=function(df_X)
   {
   new_times=base_times # 2 obs of current status, and one of previous status
   new_times$X=df_X$X
   new_times$s_time[1]=df_X$obs_wdays-df_X$trans_time # Cannot have negative values
   new_times$s_time[2]=df_X$obs_wdays-df_X$trans_time+1e-3 # So increment here
   new_times$s_time[3]=df_X$obs_wdays
   new_times$state[1]=prev_state
   new_times$state[2]=cur_state
   new_times$state[3]=1 # The censored state, not cur_state
   return(new_times)
   }

cur_state=(which(df_S$status[1] == w_status)) # status factor values
prev_state=cur_state-1

status_times=ddply(df_S, .(X), mk_X_times)

return(status_times)
}


tis=read.csv("timeInStatus.txt", as.is=TRUE)

tis$start=as.Date(tis$start, format="%Y-%m-%d") # The start column has the same value as the end column
tis$obsdate=as.Date(tis$observationDate, format="%Y-%m-%d") # Always the same data/time


# Needed to calculate business days.
# Not specified any public holidays
create.calendar(name = "US_date", weekdays = c("saturday", "sunday"))

# Number of working days
tis$obs_wdays=bizdays(tis$start, tis$obsdate, "US_date")
# tis$obs_wdays=as.numeric(tis$obsdate-tis$start)

# The base time is start.
# A state is known at some time, i.e., observationDate
# We know an extract state transition time
tis$trans_time=as.numeric(tis$obs_wdays-tis$daysInStatus)

table(tis$type)
table(tis$devTeam)
table(tis$status)
table(tis$status, tis$type)

# Major states in a sequence that looks reasonable
w_status=c("before Open", # Simplify code by having this dummy entry
		"Open",
		"Withdrawn",
		"Merged",
		"Backlog",
		"In Analysis",
		"In Refinement",
		"Ready for Development",
		"In Development",
		"Code Review",
		"Ready for Test",
		"In Testing",
		"Ready for Signoff",
		"Done")

# length(which(tis$start != tis$end))

plot(count(tis$daysInStatus), log="xy", ylim=c(1, 100))

t=count(tis$daysInStatus)
t=t[order(t$freq), decreasing=TRUE]
head(t, n=40)

plot(count(tis$obs_days), log="xy", ylim=c(1, 120))

plot(1, type="n", log="y",
	xlim=c(0, 200), ylim=c(1, 20),
	xlab="Days", ylab="Occurrences")

dum=ddply(tis, .(status), function(df) lines(count(df$trans_time)))


# Build a multi-state model

# Possible covariates: devTeam, sp, key, type

# Empty data frame, filled in using values in each row of the data (i.e., tis)
base_times=data.frame(X=rep(NA, 3),
			s_time=NA,
			state=NA)

# Some items were in their current state prior to state date, i.e., trans_time is negative
# Remove these, since we don't have a start date for them.
non_neg_trans=subset(tis, trans_time >= 0)
work_items=subset(non_neg_trans, status %in% w_status)
work_items=subset(work_items, obs_wdays > 0)

base_info=data.frame(X=work_items$X,
			devTeam=work_items$devTeam,
			key=work_items$key,
			sp=work_items$sp,
			type=work_items$type)

# Map data to the three known time/states.
# The Open state does not have a transition as such.
status_events=ddply(work_items, .(status), mk_status_hist)

# Need to transition entries in the states Withdrawn, Merged to the state Done
# ???

# Initial setup
merged_status=tmerge(base_info, work_items, id=X, tstop=obs_wdays)

# Now merge state times
merged_status=tmerge(merged_status, status_events, id=X,
			state=event(s_time, state)
			)

# This needs to be a factor for a multi-state model to be built
merged_status$state=factor(merged_status$state, 1:14, labels=w_status)

m_fit=survfit(Surv(tstart, tstop, state) ~ 1, id=X, data=merged_status)
plot(m_fit)

# Basic, catch all case
m_fit=survfit(Surv(tstop-tstart, state) ~ 1, data=merged_status)
plot(m_fit,
	xlab="Days", ylab="Probability in state")
print(m_fit)

dT_fit=survfit(Surv(tstop-tstart, state) ~ devTeam, data=merged_status)
print(dT_fit)

sp_fit=survfit(Surv(tstart, tstop, state) ~ sp, id=X, data=merged_status)
sp_fit=survfit(Surv(tstop-tstart, state) ~ sp, data=merged_status)
print(sp_fit)

mt_fit=survfit(Surv(tstop-tstart, state) ~ type, data=merged_status)
print(mt_fit)

# Ran out of iterations and did not converge
s_cox=coxph(Surv(tstart, tstop, state) ~ devTeam, id=X, data=merged_status)
summary(s_cox)


# Transition probability code that went nowhere

# Hand waving approximation to transition probabilities between states in status.
# It's a sequence of states, with a few jumps to absorbings states.
# Merged, Done and Withdrawn are absorbing states
status_tran=matrix(nrow=length(w_status), ncol=length(w_status),
		c(0, 0.01, 0.01, 0.05,    0,    0,    0,    0,    0,    0,    0,    0,    0, # Open
		  0,    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,    0, # Withdrawn
		  0,    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,    0, # Merged
		  0,    0,    0,    0, 0.05,    0,    0,    0,    0,    0,    0,    0,    0, # Backlog
		  0,    0, 0.01,    0,    0, 0.05,    0,    0,    0,    0,    0,    0,    0, # In analysis
		  0,    0, 0.01,    0,    0,    0, 0.05,    0,    0,    0,    0,    0,    0, # In Refinemen
		  0,    0,    0,    0,    0,    0,    0, 0.05,    0,    0,    0,    0,    0, # Read for Dev
		  0,    0, 0.01,    0,    0,    0,    0,    0, 0.05,    0,    0,    0,    0, # In dev
		  0,    0,    0,    0,    0,    0,    0,    0,    0, 0.05,    0,    0,    0, # Code Review
		  0,    0,    0,    0,    0,    0,    0,    0,    0,    0, 0.05,    0,    0, # Ready for T
		  0,    0,    0,    0,    0,    0,    0,    0,    0,    0,    0, 0.05,    0, # In Testing
		  0,    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,    0, 0.05, # Ready for S
		  0,    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,    0  # Done
		 )
		)



